from dataclasses import dataclass

from news_toolkit_api.db.subscription import SubscriptionPlan


@dataclass(frozen=True)
class SubscriptionCost:
    monthly: int
    annual: int


@dataclass(frozen=True)
class SubscriptionResponse:
    id: str
    name: SubscriptionPlan
    benefits: list[str]
    cost: SubscriptionCost


@dataclass(frozen=True)
class SubscriptionsResponse:
    subscriptions: list[SubscriptionResponse]



from news_toolkit_api.api.v1.response.article import ArticleResponse, ContentType
from news_toolkit_api.api.v1.response.category import CategoriesResponse
from news_toolkit_api.api.v1.response.related_articles import (
    BlockType,
    NavigateToArticleAction,
    RelatedArticleResponse,
    RelatedArticlesResponse,
)
from news_toolkit_api.api.v1.response.subscription import (
    SubscriptionCost,
    SubscriptionResponse,
    SubscriptionsResponse,
)



from abc import ABCMeta, abstractmethod
from datetime import datetime

from bs4 import BeautifulSoup

from news_toolkit_api.client import fetch_content
from news_toolkit_api.db import Article, RelatedArticle


class ArticleRepository(metaclass=ABCMeta):
    def __init__(self) -> None:
        self.html_text: str | None = None
        self.soup: BeautifulSoup | None = None

    @abstractmethod
    def build_url(self, article_id: str) -> str:
        pass

    @abstractmethod
    def parse_title(self) -> str | None:
        pass

    @abstractmethod
    def parse_content(self) -> list[str] | None:
        pass

    @abstractmethod
    def parse_category(self) -> str | None:
        pass

    @abstractmethod
    def parse_image_url(self) -> str | None:
        pass

    @abstractmethod
    def parse_auther(self) -> str | None:
        pass

    @abstractmethod
    def parse_published_at(self) -> datetime | None:
        pass

    @abstractmethod
    def parse_related_articles(self) -> list[RelatedArticle]:
        pass

    async def fetch_content(self, article_id: str) -> Article | None:
        article_url = self.build_url(article_id)
        self.html_text = await fetch_content(article_url)
        if not self.html_text:
            return None

        self.soup = BeautifulSoup(self.html_text, features="html5lib")
        if not self.soup:
            return None

        title = self.parse_title()
        content = self.parse_content()
        category = self.parse_category()
        if not title or not content or not category:
            return None

        return Article(
            article_id=article_id,
            title=title,
            content=content,
            url=article_url,
            category=category,
            image_url=self.parse_image_url(),
            auther=self.parse_auther(),
            published_at=self.parse_published_at(),
            related_articles=self.parse_related_articles(),
            is_premium=False,
            is_preview=False,
            created_at=datetime.utcnow(),
        )



from fastapi import FastAPI
from fastapi_injector import attach_injector
from injector import Injector

from news_toolkit_api.api.v1.routers import router


def create_app(injector: Injector) -> FastAPI:
    app = FastAPI()
    app.include_router(router)
    attach_injector(app, injector)
    return app


if __name__ == "__main__":
    import uvicorn

    app = create_app(Injector())
    uvicorn.run("main:app", host="0.0.0.0", port=5555, reload=True)



from dataclasses import dataclass


@dataclass(frozen=True)
class TextLeadParagraphBlock:
    text: str
    __identifier = "__text_lead_paragraph__"

    @property
    def type(self) -> str:
        return self.__identifier



import asyncio
import logging
import os

import aiohttp
from aiohttp import ClientTimeout

REQUEST_TIMEOUT = int(os.getenv("REQUEST_TIMEOUT", 60))


async def fetch_content(url: str) -> str | None:
    try:
        async with aiohttp.ClientSession(
            timeout=ClientTimeout(REQUEST_TIMEOUT)
        ) as session:
            async with session.get(url) as res:
                res.raise_for_status()
                return await res.text()
    except aiohttp.ClientError as e:
        logging.error(f"Error fetching the url {url}: {e}")
        return None
    except asyncio.TimeoutError:
        logging.error(f"Timeout while fetching the url {url}.")
        return None



from fastapi import APIRouter, BackgroundTasks, Depends
from fastapi_injector import Injected
from google.cloud import ndb

from news_toolkit_api.api.v1.db import get_client
from news_toolkit_api.api.v1.response import (
    ArticleResponse,
    BlockType,
    CategoriesResponse,
    ContentType,
    NavigateToArticleAction,
    RelatedArticleResponse,
    RelatedArticlesResponse,
    SubscriptionCost,
    SubscriptionResponse,
    SubscriptionsResponse,
)
from news_toolkit_api.background_tasks.article import background_task_article
from news_toolkit_api.db import Article, Category, Subscription
from news_toolkit_api.news_blocks import (
    ArticleIntroductionBlock,
    BannerAdContent,
    BannerAdSize,
    TextLeadParagraphBlock,
)
from news_toolkit_api.repository import ArticleRepository
from news_toolkit_api.utils import needs_update, sha3_256_hash

router = APIRouter(prefix="/api/v1")


@router.get("/articles/{article_id}")
async def get_article(
    article_id: str,
    background_tasks: BackgroundTasks,
    limit: int = 20,
    offset: int = 0,
    client: ndb.Client = Depends(get_client),
    article_repository: ArticleRepository = Injected(ArticleRepository),
):
    with client.context():
        article: Article = ndb.Key(Article, sha3_256_hash(article_id)).get()

    if not article:
        article = await background_task_article(
            client, article_repository, article_id, True
        )
        if not article:
            return {}

    if needs_update(article.created_at):
        background_tasks.add_task(
            background_task_article, client, article_repository, article_id, True
        )

    contents: ContentType = [
        ArticleIntroductionBlock(
            title=article.title,
            category=article.category,
            image_url=article.image_url,
            author=article.auther,
            published_at=article.published_at,
            is_premium=False,
        )
    ]
    for content in article.content:
        match content:
            case "ADVERTISEMENT":
                contents.append(BannerAdContent(size=BannerAdSize.large))
            case _:
                contents.append(TextLeadParagraphBlock(text=content))

    return ArticleResponse(
        title=article.title,
        content=contents[offset : offset + limit],
        url=article.url,
        is_premium=article.is_premium,
        is_preview=article.is_preview,
        total_count=len(article.content),
    )


@router.get("/articles/{article_id}/related")
async def get_related_articles(
    article_id: str,
    background_tasks: BackgroundTasks,
    limit: int = 20,
    offset: int = 0,
    client: ndb.Client = Depends(get_client),
    article_repository: ArticleRepository = Injected(ArticleRepository),
):
    with client.context():
        article: Article = ndb.Key(Article, sha3_256_hash(article_id)).get()

    if not article or needs_update(article.created_at):
        background_tasks.add_task(
            background_task_article, client, article_repository, article_id, True
        )
        return {}

    related_articles = []
    for related_article in article.related_articles[offset : offset + limit]:
        related_articles.append(
            RelatedArticleResponse(
                id=related_article.article_id,
                title=related_article.title,
                category=related_article.category,
                image_url=related_article.image_url,
                author=related_article.author,
                published_at=related_article.published_at,
                is_premium=False,
                type=BlockType.post_small,
                action=NavigateToArticleAction(
                    article_id=related_article.article_id,
                ),
            )
        )
    return RelatedArticlesResponse(
        related_articles=related_articles,
        total_count=len(article.related_articles),
    )


@router.get("/subscriptions")
async def get_subscriptions(client: ndb.Client = Depends(get_client)):
    with client.context():
        subscriptions = []
        for subscription in Subscription.query():
            subscriptions.append(
                SubscriptionResponse(
                    id=subscription.key.id(),
                    name=subscription.name,
                    benefits=subscription.benefits,
                    cost=SubscriptionCost(
                        monthly=subscription.cost.monthly,
                        annual=subscription.cost.annual,
                    ),
                )
            )

    return SubscriptionsResponse(subscriptions=subscriptions)


@router.get("/categories")
async def get_categories(client: ndb.Client = Depends(get_client)):
    with client.context():
        categories = [category.name for category in Category.query()]

    return CategoriesResponse(categories=categories)


# fake endpoints


@router.get("/search/popular")
async def get_search_popular():
    return {
        "articles": [
            {
                "id": "5c47495a-608b-4e8b-a7f0-642a02594888",
                "category": "news",
                "author": "CNN",
                "published_at": "2022-03-17T00:00:00.000",
                "image_url": "https://scitechdaily.com/images/Ear-Hearing-Concept.jpg",
                "title": "Boeing makes third attempt to launch its Starliner capsule to the ISS",
                "description": "Boeing will try yet again Thursday to send the capsule it...",
                "is_premium": False,
                "type": "__post_small__",
            }
        ],
        "topics": [],
    }


@router.get("/search/relevant")
async def get_search_relevant():
    return {
        "articles": [
            {
                "id": "b1fc2ffc-eb02-42ce-af65-79702172a987",
                "category": "news",
                "author": "Northwestern University",
                "published_at": "2022-03-11T00:00:00.000",
                "image_url": "https://scitechdaily.com/images/Ear-Hearing-Concept.jpg",
                "title": "Restoring Hearing: New Tool To Create Ear Hair Cells Lost Due to Aging or Noise",
                "description": "‘We have overcome a major hurdle’ to restore hearing,...",
                "is_premium": False,
                "type": "__post_small__",
            }
        ],
        "topics": [],
    }


@router.get("/users/me")
async def get_me():
    return {
        "user": {
            "id": "2e99887d-d672-4b96-ad6a-123c1c7fa3fa",
            "subscription_plan": "premium",
        }
    }



from news_toolkit_api.db.article import Article, RelatedArticle
from news_toolkit_api.db.category import Category
from news_toolkit_api.db.subscription import (
    Subscription,
    SubscriptionCost,
    SubscriptionPlan,
)



from google.cloud import ndb


class RelatedArticle(ndb.Model):
    article_id = ndb.StringProperty(required=True)
    title = ndb.StringProperty(required=True)
    category = ndb.StringProperty(required=True)
    image_url = ndb.StringProperty(required=True)
    auther = ndb.StringProperty()
    published_at = ndb.DateTimeProperty()


class Article(ndb.Model):
    article_id = ndb.StringProperty(required=True)
    title = ndb.StringProperty(required=True)
    content = ndb.TextProperty(repeated=True)
    url = ndb.StringProperty(required=True)
    category = ndb.StringProperty(required=True)
    image_url = ndb.StringProperty()
    auther = ndb.StringProperty()
    published_at = ndb.DateTimeProperty()
    is_premium = ndb.BooleanProperty(default=False)
    is_preview = ndb.BooleanProperty(default=False)
    related_articles = ndb.StructuredProperty(RelatedArticle, repeated=True)
    created_at = ndb.DateTimeProperty(required=True)



from google.cloud import ndb

from news_toolkit_api.db.article import Article
from news_toolkit_api.repository.article import ArticleRepository
from news_toolkit_api.utils import sha3_256_hash


async def background_task_article(
    client: ndb.Client,
    article_repository: ArticleRepository,
    article_id: str,
    fetch_related_article: bool = True,
):
    article = await article_repository.fetch_content(article_id)
    if not article:
        return None

    with client.context():
        article.key = ndb.Key(Article, sha3_256_hash(article.article_id))
        article.put()

    if fetch_related_article:
        for related_article in article.related_articles:
            await background_task_article(
                client, article_repository, related_article.article_id, False
            )

    return article



from dataclasses import dataclass
from datetime import datetime
from enum import Enum


@dataclass(frozen=True)
class NavigateToArticleAction:
    article_id: str
    __identifier = "__navigate_to_article__"

    def type(self) -> str:
        return self.__identifier


class BlockType(Enum):
    post_small = "__post_small__"


@dataclass(frozen=True)
class RelatedArticleResponse:
    id: str
    title: str
    category: str
    image_url: str
    author: str | None
    published_at: datetime | None
    is_premium: bool
    type: BlockType
    action: NavigateToArticleAction


@dataclass(frozen=True)
class RelatedArticlesResponse:
    related_articles: list[RelatedArticleResponse]
    total_count: int
